// Copyright (C) 2008 Cory Strope <corystrope@gmail.com>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later
// version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef _TREE_H_
#define _TREE_H_

#include <string>
#include <cstring>
#include <vector>
#include <list>
#include <fstream>

using namespace std;

#define MAX_NAME_LEN 256
#define MAX_INDEL_SIZE 100

extern int total_motif_positions;
extern int correct_motif_positions;
extern int clone_iSGv1;
extern int num_template_violations;

// Forward definitions?
class TTree;
class TNode;
class inClade;
class seqGenOptions;
class motifSite;
class varSite;
class inTree;
class siteRegEx;
class inMotif;

class TNode {
public:
	// Variables
	TNode *branch0, *branch1, *branch2, *anc;
	double length0, length1, length2, param;
	int tipNo;
	string sequence;
	vector<motifSite*> myMotif;
	vector<short> invariable;	
	vector<double> gammaRates;	
	vector<short> categories;
	int  indel_num;
	double branch1_time_relative_length, branch2_time_relative_length, branch0_time_relative_length;
	double branch1_max_path, branch2_max_path;
	double DistanceFromRoot;
	int numEventsToSimulate;
	double BL_step_size;
	double atEpochTime, example_epoch_time;
	vector<bool> bipartition;
	string clade_label;
	inClade *nodeEnv;
	int mytipNo;
	varSite *one_site_varSite;
	varSite *unconstrained_varSite;
	list<varSite*> variable_region_list;

	// Functions:
	bool varSiteOverlap();
	bool isVarSite(varSite *chk_varSite);
	TNode(TTree *tree);
	varSite *Generic_varSite();
	void calcMotifAccuracy();
	void report_varSites();
	void Inherit_varSites(list<varSite*> ancvarsite);
	void addGeneral_varSites();
	void InheritMotifSites();
	void clearMotifSites();
	void report(TTree *tree = NULL);
	void postProcess(); 	// A routine to set all tracking pointers for sequence item.
	void dumpBetweenSitePointers(string root_sequence);
	void Print_Active_Sites(string root_sequence);
	void FULL_REPORT();
	void setFirstMotifPos();
	void setRemainingPos();
	void setInvariableArrayPos();
	void Remove_Objects(bool root_node = false);
	void setZeroSites();
	void report_on_sites();
	string output_sequence(string::iterator start, string::iterator end);
	size_t findFirstConstrained(size_t fromSite);
	size_t findForwardNumberOfConstrained(size_t fromSite, size_t numSites);
	size_t findBackwardNumberOfConstrained(size_t fromSite, size_t numSites);

};

class TTree {
public:
	// Variables
	bool codons_only;
	int rooted, lengths;
	TNode *root; 
	list<TNode*> nodeList;
	int numTips, numNodes;
	double totalLength;
	vector<string> names;
	vector<TNode*> tips;
	int root_numSites;
	double epoch_step_size;
	double trs_example_step_size;
	list<inClade*> treeEnv;
	// Global arrays.
	vector<char> global_alignment_action;
	vector<int>  global_alignment_fromAnc;
	vector<int>  global_alignment_indelNo;

	// Functions:
	TNode *ReadTip(string tree_str, int *pos, char ch, TTree *tree, bool first_partition);
	TNode *ReadNode(string tree_str, int *pos, TTree *tree, int detectPolytomies, bool first_partition);
	TNode *ReadBranch(string tree_str, int *pos, TTree *tree, bool first_partition);
	TNode *AddNode();
	inClade *AddClade(string& label);
	void constructRootMotifSites(inTree *metree, string &root_sequence, string &invariable);
	void ReadTree(string str, int *pos, TTree *tree, bool first_partition);
	void InitTree();
	void report_clades();
	void Place_Motif(list<siteRegEx*>& motif, char mark, list<inMotif*>& placed_motifs, inMotif* thisMotif, string &root_sequence, size_t start_site);

	list<inMotif*> DrawMotifs(string& root_sequence, double proportion_motifs);
	TTree();
};

#endif /* _TREE_H_ */

